"""
slow orbit feedback package
~~~~~~~~~~~~~~~~

pyAML

"""

__title__ = "pyAML"
__description__ = "Python Accelerator Middle Layer"
__url__ = "https://github.com/python-accelerator-middle-layer/pyaml"
__version__ = "0.0.0"
__author__ = "pyAML collaboration"
__author_email__ = ""

__all__ = [__version__]
