"""
PyAML global constants
"""

# Horizontal kick sign convention for both kick an kick angle
HORIZONTAL_KICK_SIGN: float = -1.0

# Action identifier for callback in response matrix measurement
ACTION_APPLY: int = 0
ACTION_RESTORE: int = 1
